# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg
from oslo_messaging.rpc import dispatcher as rpc
from oslo_utils import uuidutils
import six

from senlin.common import consts
from senlin.common import exception as exc
from senlin.common.i18n import _
from senlin.common import scaleutils as su
from senlin.common import utils as common_utils
from senlin.engine.actions import base as am
from senlin.engine import dispatcher
from senlin.engine import node as nm
from senlin.engine import service
from senlin.objects import action as ao
from senlin.objects import base as obj_base
from senlin.objects import cluster as co
from senlin.objects import cluster_policy as cpo
from senlin.objects import node as no
from senlin.objects import profile as po
from senlin.objects import receiver as ro
from senlin.objects.requests import clusters as orco
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class ClusterTest(base.SenlinTestCase):

    def setUp(self):
        super(ClusterTest, self).setUp()

        self.ctx = utils.dummy_context(project='cluster_test_project')
        self.eng = service.EngineService('host-a', 'topic-a')

    @mock.patch.object(co.Cluster, 'count_all')
    def test_check_cluster_quota(self, mock_count):
        mock_count.return_value = 10
        cfg.CONF.set_override('max_clusters_per_project', 11)

        res = self.eng.check_cluster_quota(self.ctx)

        self.assertIsNone(res)
        mock_count.assert_called_once_with(self.ctx)

    @mock.patch.object(co.Cluster, 'count_all')
    def test_check_cluster_quota_failed(self, mock_count):
        mock_count.return_value = 11
        cfg.CONF.set_override('max_clusters_per_project', 11)

        ex = self.assertRaises(exc.Forbidden,
                               self.eng.check_cluster_quota, self.ctx)
        self.assertEqual("You are not authorized to complete this "
                         "operation.",
                         six.text_type(ex))

    def _prepare_request(self, req):
        mock_cls = self.patchobject(obj_base.SenlinObject,
                                    'obj_class_from_name')
        req.update({'senlin_object.name': 'RequestClass',
                    'senlin_object.version': '1.0'})
        req_base = mock.Mock()
        mock_cls.return_value = req_base
        req_obj = mock.Mock()
        for k, v in req.items():
            setattr(req_obj, k, v)
        req_base.obj_from_primitive.return_value = req_obj

    @mock.patch.object(co.Cluster, 'get_all')
    def test_cluster_list(self, mock_get):
        x_obj_1 = mock.Mock()
        x_obj_1.to_dict.return_value = {'k': 'v1'}
        x_obj_2 = mock.Mock()
        x_obj_2.to_dict.return_value = {'k': 'v2'}
        mock_get.return_value = [x_obj_1, x_obj_2]
        req = orco.ClusterListRequest(project_safe=True)

        result = self.eng.cluster_list(self.ctx, req.obj_to_primitive())

        self.assertEqual([{'k': 'v1'}, {'k': 'v2'}], result)
        mock_get.assert_called_once_with(self.ctx, project_safe=True)

    @mock.patch.object(co.Cluster, 'get_all')
    def test_cluster_list_with_params(self, mock_get):
        mock_get.return_value = []
        marker = uuidutils.generate_uuid()
        req = {
            'limit': 10,
            'marker': marker,
            'name': ['test_cluster'],
            'status': ['ACTIVE'],
            'sort': 'name:asc',
            'project_safe': True
        }
        self._prepare_request(req)

        result = self.eng.cluster_list(self.ctx, req)

        self.assertEqual([], result)
        mock_get.assert_called_once_with(
            self.ctx, limit=10, marker=marker, sort='name:asc',
            filters={'name': ['test_cluster'], 'status': ['ACTIVE']},
            project_safe=True)

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, "create")
    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_create(self, notify, mock_profile, mock_cluster,
                            mock_action, mock_check, mock_quota):
        x_profile = mock.Mock(id='PROFILE_ID')
        mock_profile.return_value = x_profile
        x_cluster = mock.Mock(id='12345678ABC')
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        mock_check.return_value = None
        mock_quota.return_value = None
        req = orco.ClusterCreateRequestBody(name='C1', profile_id='PROFILE',
                                            desired_capacity=3)

        # do it
        result = self.eng.cluster_create(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_profile.assert_called_once_with(self.ctx, 'PROFILE')
        mock_check.assert_called_once_with(None, 3, None, None, True)
        mock_cluster.assert_called_once_with(
            self.ctx,
            dict(name='C1', desired_capacity=3, profile_id='PROFILE_ID',
                 min_size=0, max_size=-1, timeout=3600, metadata={},
                 dependents={}, data={}, next_index=1, status='INIT',
                 config={},
                 status_reason='Initializing', user=self.ctx.user,
                 project=self.ctx.project, domain=self.ctx.domain))
        mock_action.assert_called_once_with(
            self.ctx,
            '12345678ABC', 'CLUSTER_CREATE',
            name='cluster_create_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, "create")
    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_create_desired_null(self, notify, mock_profile,
                                         mock_cluster, mock_action,
                                         mock_check, mock_quota):
        x_profile = mock.Mock(id='PROFILE_ID')
        mock_profile.return_value = x_profile
        x_cluster = mock.Mock(id='12345678ABC')
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        mock_check.return_value = None
        mock_quota.return_value = None
        req = orco.ClusterCreateRequestBody(name='C1', profile_id='PROFILE',
                                            min_size=1, max_size=5,
                                            config={'k1': 'v1'})

        # do it
        result = self.eng.cluster_create(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_profile.assert_called_once_with(self.ctx, 'PROFILE')
        mock_check.assert_called_once_with(None, 1, 1, 5, True)
        mock_cluster.assert_called_once_with(
            self.ctx,
            dict(name='C1', desired_capacity=1, profile_id='PROFILE_ID',
                 min_size=1, max_size=5, timeout=3600, metadata={},
                 dependents={}, data={}, next_index=1, status='INIT',
                 config={'k1': 'v1'},
                 status_reason='Initializing', user=self.ctx.user,
                 project=self.ctx.project, domain=self.ctx.domain))
        mock_action.assert_called_once_with(
            self.ctx,
            '12345678ABC', 'CLUSTER_CREATE',
            name='cluster_create_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    def test_cluster_create_exceeding_quota(self, mock_quota):
        mock_quota.side_effect = exc.Forbidden()
        req = {'profile_id': 'PROFILE', 'name': 'CLUSTER'}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, req)

        self.assertEqual(exc.Forbidden, ex.exc_info[0])
        self.assertEqual("You are not authorized to complete this "
                         "operation.",
                         six.text_type(ex.exc_info[1]))
        mock_quota.assert_called_once_with(self.ctx)

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_cluster_create_duplicate_name(self, mock_get, mock_quota):
        cfg.CONF.set_override('name_unique', True)
        mock_quota.return_value = None
        mock_get.return_value = mock.Mock()
        req = {'profile_id': 'PROFILE', 'name': 'CLUSTER'}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual(_("a cluster named 'CLUSTER' already exists."),
                         six.text_type(ex.exc_info[1]))
        mock_get.assert_called_once_with(self.ctx, 'CLUSTER')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(po.Profile, 'find')
    def test_cluster_create_profile_not_found(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.side_effect = exc.ResourceNotFound(type='profile',
                                                     id='Bogus')
        req = {'profile_id': 'Bogus', 'name': 'CLUSTER'}
        self._prepare_request(req)
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The specified profile 'Bogus' could not "
                         "be found.", six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(su, 'check_size_params')
    def test_cluster_create_failed_checking(self, mock_check, mock_find,
                                            mock_quota):
        mock_quota.return_value = None
        mock_find.return_value = mock.Mock()
        mock_check.return_value = 'INVALID'
        req = {'profile_id': 'PROFILE', 'name': 'CLUSTER'}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("INVALID.", six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'PROFILE')

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_get(self, mock_find):
        x_cluster = mock.Mock()
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_find.return_value = x_cluster
        req = orco.ClusterGetRequest(identity='C1')

        result = self.eng.cluster_get(self.ctx, req.obj_to_primitive())

        self.assertEqual({'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_get_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = {'identity': 'CLUSTER'}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_get,
                               self.ctx, req)
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update(self, notify, mock_find, mock_profile,
                            mock_action):
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              profile_id='OLD_PROFILE',
                              metadata={'A': 'B'})
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_find.return_value = x_cluster
        old_profile = mock.Mock(type='FAKE_TYPE', id='ID_OLD')
        new_profile = mock.Mock(type='FAKE_TYPE', id='ID_NEW')
        mock_profile.side_effect = [old_profile, new_profile]
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterUpdateRequest(identity='FAKE_ID', name='new_name',
                                        profile_id='NEW_PROFILE',
                                        metadata={'B': 'A'}, timeout=120,
                                        config={'k1': 'v1'})

        # do it
        result = self.eng.cluster_update(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_PROFILE'),
            mock.call(self.ctx, 'NEW_PROFILE'),
        ])
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                'new_profile_id': 'ID_NEW',
                'metadata': {
                    'B': 'A',
                },
                'timeout': 120,
                'name': 'new_name',
                'config': {
                    'k1': 'v1',
                },
            }
        )

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_update_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = {'identity': 'Bogus', 'name': 'new-name'}
        self._prepare_request(req)
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, req)
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_update_cluster_bad_status(self, mock_find):
        x_cluster = mock.Mock(status='ERROR')
        mock_find.return_value = x_cluster
        req = {'identity': 'CLUSTER', 'name': 'new-name'}
        self._prepare_request(req)

        self.assertEqual(consts.CS_ERROR, x_cluster.status)
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, req)

        self.assertEqual(exc.FeatureNotSupported, ex.exc_info[0])
        self.assertEqual('Updating a cluster in error state is not supported.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')

    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_update_profile_not_found(self, mock_find, mock_profile):
        mock_find.return_value = mock.Mock(status='ACTIVE',
                                           profile_id='OLD_ID')
        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE', id='OLD_ID'),
            exc.ResourceNotFound(type='profile', id='Bogus')
        ]
        req = orco.ClusterUpdateRequest(identity='CLUSTER', profile_id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The specified profile 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_ID'),
            mock.call(self.ctx, 'Bogus'),
        ])

    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_update_diff_profile_type(self, mock_find, mock_profile):
        x_obj = mock.Mock(status='ACTIVE', profile_id='OLD_ID')
        mock_find.return_value = x_obj
        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE', id='OLD_ID'),
            mock.Mock(type='DIFF_TYPE', id='NEW_ID'),
        ]
        req = orco.ClusterUpdateRequest(identity='CLUSTER',
                                        profile_id='NEW_PROFILE')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_ID'),
            mock.call(self.ctx, 'NEW_PROFILE'),
        ])

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update_same_profile(self, notify, mock_find,
                                         mock_profile, mock_action):
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              profile_id='OLD_PROFILE')
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_find.return_value = x_cluster
        old_profile = mock.Mock(type='FAKE_TYPE', id='ID_OLD')
        new_profile = mock.Mock(type='FAKE_TYPE', id='ID_OLD')
        mock_profile.side_effect = [old_profile, new_profile]
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterUpdateRequest(identity='FAKE_ID', name='NEW_NAME',
                                        profile_id='NEW_PROFILE')

        # do it
        result = self.eng.cluster_update(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_PROFILE'),
            mock.call(self.ctx, 'NEW_PROFILE'),
        ])
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                # Note profile_id is not shown in the inputs
                'name': 'NEW_NAME',
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update_same_metadata(self, notify, mock_find,
                                          mock_action):
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              metadata={'K': 'V'})
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterUpdateRequest(identity='FAKE_ID', name='NEW_NAME',
                                        metadata={'K': 'V'})

        # do it
        result = self.eng.cluster_update(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            status=am.Action.READY,
            cause=consts.CAUSE_RPC,
            inputs={
                # Note metadata is not included in the inputs
                'name': 'NEW_NAME',
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update_same_timeout(self, notify, mock_find,
                                         mock_action):
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              timeout=10)
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        x_cluster.timeout = 10
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterUpdateRequest(identity='FAKE_ID', name='NEW_NAME',
                                        timeout=10)

        # do it
        result = self.eng.cluster_update(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            status=am.Action.READY,
            cause=consts.CAUSE_RPC,
            inputs={
                # Note timeout is not included in the inputs
                'name': 'NEW_NAME',
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update_same_name(self, notify, mock_find,
                                      mock_action):
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              name='OLD_NAME', timeout=10)
        x_cluster.name = 'OLD_NAME'
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterUpdateRequest(identity='FAKE_ID', name='OLD_NAME',
                                        timeout=100)

        # do it
        result = self.eng.cluster_update(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            status=am.Action.READY,
            cause=consts.CAUSE_RPC,
            inputs={
                # Note name is not included in the inputs
                'timeout': 100,
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_update_all_property_same(self, mock_find):
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              name='OLD_NAME', timeout=10)
        x_cluster.name = 'OLD_NAME'
        x_cluster.timeout = 10
        mock_find.return_value = x_cluster

        # Notice that name and timeout are all not changed.
        req = orco.ClusterUpdateRequest(identity='CLUSTER', name='OLD_NAME',
                                        timeout=10)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual('', six.text_type(ex))

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_update_no_property_updated(self, mock_find):
        x_cluster = mock.Mock(status='ACTIVE', profile_id='OLD_ID')
        mock_find.return_value = x_cluster
        req = orco.ClusterUpdateRequest(identity='CLUSTER')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual('', six.text_type(ex))

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = {'identity': 'Bogus', 'nodes': ['n1', 'n2']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_add_nodes(self, notify, mock_find, mock_node,
                               mock_profile, mock_action, mock_check):
        x_cluster = mock.Mock(id='12345678AB', profile_id='FAKE_ID',
                              desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        x_node_1 = mock.Mock(id='NODE1', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_ID_1')
        x_node_2 = mock.Mock(id='NODE2', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_ID_1')
        mock_node.side_effect = [x_node_1, x_node_2]
        mock_action.return_value = 'ACTION_ID'
        mock_check.return_value = None
        req = orco.ClusterAddNodesRequest(identity='C1',
                                          nodes=['NODE_A', 'NODE_B'])

        result = self.eng.cluster_add_nodes(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'NODE_A'),
            mock.call(self.ctx, 'NODE_B'),
        ])
        mock_check.assert_called_once_with(x_cluster, 6, strict=True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', consts.CLUSTER_ADD_NODES,
            name='cluster_add_nodes_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'nodes': ['NODE1', 'NODE2']},
        )
        self.assertEqual(3, mock_profile.call_count)
        notify.assert_called_once_with()

    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_nodes_not_found(self, mock_find, mock_node,
                                               mock_profile):
        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.side_effect = exc.ResourceNotFound(type='node', id='NODE1')
        req = {'identity': 'CLUSTER', 'nodes': ['NODE1']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Nodes not found: ['NODE1'].",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_called_once_with(self.ctx, 'FAKE_ID',
                                             project_safe=True)
        mock_node.assert_called_once_with(self.ctx, 'NODE1')

    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_bad_status(self, mock_find, mock_node,
                                          mock_profile):
        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.return_value = mock.Mock(
            id='NODE2', cluster_id='', status='ERROR')
        req = {'identity': 'CLUSTER', 'nodes': ['NODE2']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Nodes are not ACTIVE: ['NODE2'].",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        self.assertEqual(2, mock_profile.call_count)
        mock_node.assert_called_once_with(self.ctx, 'NODE2')

    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_node_already_owned(self, mock_find,
                                                  mock_node, mock_profile):

        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.return_value = mock.Mock(id='NODE3', status='ACTIVE',
                                           cluster_id='OTHER')
        req = {'identity': 'CLUSTER', 'nodes': ['NODE3']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Nodes ['NODE3'] already owned by some cluster.",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        self.assertEqual(2, mock_profile.call_count)
        mock_node.assert_called_once_with(self.ctx, 'NODE3')

    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_node_profile_type_not_match(
            self, mock_find, mock_node, mock_profile):

        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE_1'),
            mock.Mock(type='FAKE_TYPE_2'),
        ]
        mock_node.return_value = mock.Mock(id='NODE4', status='ACTIVE',
                                           cluster_id='', profile_id='DIFF')
        req = {'identity': 'CLUSTER', 'nodes': ['NODE4']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Profile type of nodes ['NODE4'] does not "
                         "match that of the cluster.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'FAKE_ID', project_safe=True),
            mock.call(self.ctx, 'DIFF', project_safe=True),
        ])
        mock_node.assert_called_once_with(self.ctx, 'NODE4')

    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_mult_err(self, mock_find, mock_node,
                                        mock_profile):
        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.return_value = mock.Mock(id='NODE2', status='ERROR')
        req = {'identity': 'CLUSTER', 'nodes': ['NODE2']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        msg1 = _("Nodes ['NODE2'] already owned by some cluster.")
        msg2 = _("Nodes are not ACTIVE: ['NODE2'].")
        self.assertIn(msg1, six.text_type(ex.exc_info[1]))
        self.assertIn(msg2, six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        self.assertEqual(2, mock_profile.call_count)
        mock_node.assert_called_once_with(self.ctx, 'NODE2')

    @mock.patch.object(po.Profile, 'get')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_add_nodes_failed_checking(self, mock_find, mock_node,
                                               mock_check, mock_profile):
        x_cluster = mock.Mock(id='12345678AB', profile_id='FAKE_PROFILE',
                              desired_capacity=2)
        mock_find.return_value = x_cluster
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        x_node_1 = mock.Mock(id='NODE1', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_PROFILE_1')
        x_node_2 = mock.Mock(id='NODE2', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_PROFILE_2')
        mock_node.side_effect = [x_node_1, x_node_2]
        mock_check.return_value = 'Failed size checking.'
        req = {'identity': 'C1', 'nodes': ['NODE_A', 'NODE_B']}
        self._prepare_request(req)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, req)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Failed size checking.",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'FAKE_PROFILE', project_safe=True),
            mock.call(self.ctx, 'FAKE_PROFILE_1', project_safe=True),
            mock.call(self.ctx, 'FAKE_PROFILE_2', project_safe=True),
        ])
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'NODE_A'),
            mock.call(self.ctx, 'NODE_B'),
        ])
        mock_check.assert_called_once_with(x_cluster, 4, strict=True)

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_del_nodes(self, notify, mock_find, mock_node,
                               mock_action, mock_check):
        x_cluster = mock.Mock(id='1234', desired_capacity=2)
        mock_find.return_value = x_cluster
        mock_node.return_value = mock.Mock(id='NODE2', cluster_id='1234',
                                           dependents={})
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterDelNodesRequest(identity='CLUSTER', nodes=['NODE1'])

        result = self.eng.cluster_del_nodes(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE1')
        mock_check.asset_called_once_with(x_cluster, 1, strict=True)
        mock_action.assert_called_once_with(
            self.ctx, '1234', consts.CLUSTER_DEL_NODES,
            name='cluster_del_nodes_1234',
            status=am.Action.READY,
            cause=consts.CAUSE_RPC,
            inputs={
                'count': 1,
                'candidates': ['NODE2'],
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = orco.ClusterDelNodesRequest(identity='Bogus', nodes=['NODE1'])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_node_not_found(self, mock_find, mock_node):
        mock_find.return_value = mock.Mock()
        mock_node.side_effect = exc.ResourceNotFound(type='node', id='NODE1')
        req = orco.ClusterDelNodesRequest(identity='CLUSTER', nodes=['NODE1'])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertIn("Nodes not found",
                      six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE1')

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_have_containers(self, mock_cluster, mock_node):
        mock_cluster.return_value = mock.Mock(id='CLUSTER1')
        dependents = {'nodes': ['container1']}
        node = mock.Mock(id='NODE1', dependents=dependents,
                         cluster_id='CLUSTER1')
        mock_node.return_value = node
        req = orco.ClusterDelNodesRequest(identity='CLUSTER', nodes=['NODE1'])
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())
        self.assertEqual(exc.ResourceInUse, ex.exc_info[0])
        message = _("nodes ['NODE1'] are depended by other nodes, so can't be "
                    "deleted or become orphan nodes")
        self.assertIn(message, six.text_type(ex.exc_info[1]))

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_node_in_other_cluster(self, mock_find,
                                                     mock_node):
        mock_find.return_value = mock.Mock(id='1234')
        mock_node.return_value = mock.Mock(id='NODE2', cluster_id='5678')
        req = orco.ClusterDelNodesRequest(identity='CLUSTER', nodes=['NODE2'])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Nodes not members of specified cluster: ['NODE2'].",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE2')

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_mult_errors(self, mock_find, mock_node):
        mock_find.return_value = mock.Mock(id='1234')
        mock_node.side_effect = [mock.Mock(id='NODE1', cluster_id='5678'),
                                 exc.ResourceNotFound(type='node', id='NODE2')]
        req = orco.ClusterDelNodesRequest(identity='CLUSTER',
                                          nodes=['NODE1', 'NODE2'])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        msg1 = _("Nodes not found:")
        msg2 = _("Nodes not members of specified cluster: ['NODE1'].")
        self.assertIn(msg1, six.text_type(ex.exc_info[1]))
        self.assertIn(msg2, six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        self.assertEqual(2, mock_node.call_count)

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_orphan_nodes(self, mock_find, mock_node):
        mock_find.return_value = mock.Mock(id='1234')
        mock_node.return_value = mock.Mock(id='NODE3', cluster_id='')
        req = orco.ClusterDelNodesRequest(identity='CLUSTER', nodes=['NODE3'])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Nodes not members of specified cluster: ['NODE3'].",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE3')

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_del_nodes_failed_checking(self, mock_find, mock_node,
                                               mock_check):
        x_cluster = mock.Mock(id='1234', desired_capacity=2)
        mock_find.return_value = x_cluster
        mock_node.return_value = mock.Mock(id='NODE2', cluster_id='1234',
                                           dependents={})
        mock_check.return_value = 'Failed size checking.'
        req = orco.ClusterDelNodesRequest(identity='CLUSTER', nodes=['NODE3'])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Failed size checking.",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE3')
        mock_check.assert_called_once_with(x_cluster, 1, strict=True)

    @mock.patch.object(no.Node, 'count_by_cluster')
    @mock.patch.object(su, 'calculate_desired')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_resize_exact_capacity(self, mock_find, mock_action,
                                           notify, mock_check, mock_calc,
                                           mock_count):
        x_cluster = mock.Mock(id='12345678ABCDEFGH')
        mock_find.return_value = x_cluster
        mock_count.return_value = 3
        mock_calc.return_value = 5
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.EXACT_CAPACITY,
            number=5
        )

        res = self.eng.cluster_resize(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_calc.assert_called_once_with(3, consts.EXACT_CAPACITY, 5, None)
        mock_check.assert_called_once_with(x_cluster, 5, None, None, True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_RESIZE,
            name='cluster_resize_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                consts.ADJUSTMENT_TYPE: consts.EXACT_CAPACITY,
                consts.ADJUSTMENT_NUMBER: 5,
                consts.ADJUSTMENT_MIN_SIZE: None,
                consts.ADJUSTMENT_MAX_SIZE: None,
                consts.ADJUSTMENT_MIN_STEP: None,
                consts.ADJUSTMENT_STRICT: True
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(no.Node, 'count_by_cluster')
    @mock.patch.object(su, 'calculate_desired')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_resize_change_in_capacity(self, mock_find, mock_action,
                                               notify, mock_check, mock_calc,
                                               mock_count):
        x_cluster = mock.Mock(id='12345678ABCDEFGH')
        mock_find.return_value = x_cluster
        mock_count.return_value = 2
        mock_calc.return_value = 7
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.CHANGE_IN_CAPACITY,
            number=5
        )

        res = self.eng.cluster_resize(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_calc.assert_called_once_with(2, consts.CHANGE_IN_CAPACITY, 5,
                                          None)
        mock_check.assert_called_once_with(x_cluster, 7, None, None, True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_RESIZE,
            name='cluster_resize_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                consts.ADJUSTMENT_TYPE: consts.CHANGE_IN_CAPACITY,
                consts.ADJUSTMENT_NUMBER: 5,
                consts.ADJUSTMENT_MIN_SIZE: None,
                consts.ADJUSTMENT_MAX_SIZE: None,
                consts.ADJUSTMENT_MIN_STEP: None,
                consts.ADJUSTMENT_STRICT: True
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(no.Node, 'count_by_cluster')
    @mock.patch.object(su, 'calculate_desired')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_resize_change_in_percentage(self, mock_find, mock_action,
                                                 notify, mock_check,
                                                 mock_calc, mock_count):
        x_cluster = mock.Mock(id='12345678ABCDEFGH')
        mock_find.return_value = x_cluster
        mock_count.return_value = 10
        mock_calc.return_value = 8
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.CHANGE_IN_PERCENTAGE,
            number=15.81
        )

        res = self.eng.cluster_resize(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_calc.assert_called_once_with(10, consts.CHANGE_IN_PERCENTAGE,
                                          15.81, None)
        mock_check.assert_called_once_with(x_cluster, 8, None, None, True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_RESIZE,
            name='cluster_resize_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                consts.ADJUSTMENT_TYPE: consts.CHANGE_IN_PERCENTAGE,
                consts.ADJUSTMENT_NUMBER: 15.81,
                consts.ADJUSTMENT_MIN_SIZE: None,
                consts.ADJUSTMENT_MAX_SIZE: None,
                consts.ADJUSTMENT_MIN_STEP: None,
                consts.ADJUSTMENT_STRICT: True
            },
        )
        notify.assert_called_once_with()

    def test_cluster_resize_type_missing_number(self):
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.EXACT_CAPACITY
        )

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Missing number value for size adjustment.",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_number_without_type(self):
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            number=10
        )

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, req.obj_to_primitive())
        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Missing adjustment_type "
                         "value for size adjustment.",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_number_for_exact_capacity(self):
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.EXACT_CAPACITY,
            number=-5
        )

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The 'number' must be positive integer for "
                         "adjustment type 'EXACT_CAPACITY'.",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_resize_cluster_not_found(self, mock_find):
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.EXACT_CAPACITY,
            number=10
        )
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='CLUSTER')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, req.obj_to_primitive())

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'CLUSTER' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(no.Node, 'count_by_cluster')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_resize_failing_size_check(self, mock_find, mock_count,
                                               mock_check):
        x_cluster = mock.Mock(id='CID')
        mock_find.return_value = x_cluster
        mock_count.return_value = 5
        mock_check.return_value = 'size check.'
        req = orco.ClusterResizeRequest(
            identity='CLUSTER',
            adjustment_type=consts.EXACT_CAPACITY,
            number=5
        )

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, req.obj_to_primitive())

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_count.assert_called_once_with(self.ctx, 'CID')
        mock_check.assert_called_once_with(x_cluster, 5, None, None, True)
        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("size check.",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_out(self, mock_find, mock_check, mock_action,
                               notify):
        x_cluster = mock.Mock(id='12345678ABCDEFGH', desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterScaleOutRequest(identity='CLUSTER', count=1)

        result = self.eng.cluster_scale_out(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 5)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_SCALE_OUT,
            name='cluster_scale_out_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'count': 1},
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_out_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = orco.ClusterScaleOutRequest(identity='Bogus', count=1)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_out_count_is_none(self, mock_find, mock_action,
                                             notify):
        mock_find.return_value = mock.Mock(id='12345678ABCDEFGH',
                                           desired_capacity=4)
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterScaleOutRequest(identity='CLUSTER')

        result = self.eng.cluster_scale_out(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_SCALE_OUT,
            name='cluster_scale_out_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_out_count_zero(self, mock_find):
        mock_find.return_value = mock.Mock(desired_capacity=4)
        req = orco.ClusterScaleOutRequest(identity='CLUSTER', count=0)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Count for scale-out request cannot be 0.",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_out_failed_size_check(self, mock_find, mock_check):
        x_cluster = mock.Mock(desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = 'size limit'
        req = orco.ClusterScaleOutRequest(identity='CLUSTER', count=2)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("size limit.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 6)

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_in(self, mock_find, mock_check, mock_action,
                              notify):
        x_cluster = mock.Mock(id='12345678ABCD', desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterScaleInRequest(identity='CLUSTER', count=2)

        result = self.eng.cluster_scale_in(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 2)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCD', consts.CLUSTER_SCALE_IN,
            name='cluster_scale_in_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'count': 2},
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_in_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = orco.ClusterScaleInRequest(identity='Bogus', count=2)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_in_count_is_none(self, mock_find, mock_action,
                                            notify):
        mock_find.return_value = mock.Mock(id='FOO', desired_capacity=4)
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterScaleInRequest(identity='CLUSTER')

        result = self.eng.cluster_scale_in(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_action.assert_called_once_with(
            self.ctx, 'FOO', consts.CLUSTER_SCALE_IN,
            name='cluster_scale_in_FOO',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_in_count_zero(self, mock_find):
        mock_find.return_value = mock.Mock(desired_capacity=4)
        req = orco.ClusterScaleInRequest(identity='CLUSTER', count=0)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Count for scale-in request cannot be 0.",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_scale_in_failed_size_check(self, mock_find, mock_check):
        x_cluster = mock.Mock(desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = 'size limit'
        req = orco.ClusterScaleInRequest(identity='FAKE_CLUSTER', count=2)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("size limit.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 2)

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_check(self, notify, mock_find, mock_action):
        x_cluster = mock.Mock(id='CID', user='USER', project='PROJECT')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterCheckRequest(identity='C1', params={'foo': 'bar'})

        res = self.eng.cluster_check(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_CHECK,
            name='cluster_check_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'foo': 'bar'},
        )
        notify.assert_called_once_with()

    @mock.patch.object(ao.Action, 'delete_by_target')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_check_with_delete(self, notify, mock_find, mock_action,
                                       mock_delete):
        x_cluster = mock.Mock(id='CID', user='USER', project='PROJECT')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterCheckRequest(identity='C1',
                                       params={'delete_check_action': True})

        res = self.eng.cluster_check(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_delete.assert_called_once_with(
            self.ctx, 'CID', action=['CLUSTER_CHECK'],
            status=['SUCCEEDED', 'FAILED']
        )
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_CHECK,
            name='cluster_check_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'delete_check_action': True},
        )
        notify.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_check_user_is_none(self, notify, mock_find, mock_action):
        x_cluster = mock.Mock(id='CID', project='PROJECT')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterCheckRequest(identity='C1')

        result = self.eng.cluster_check(self.ctx, req.obj_to_primitive())

        self.assertIsNotNone(x_cluster.user)
        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_CHECK,
            name='cluster_check_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_check_project_is_none(self, notify, mock_find,
                                           mock_action):
        x_cluster = mock.Mock(id='CID', user='USER')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterCheckRequest(identity='C1')

        result = self.eng.cluster_check(self.ctx, req.obj_to_primitive())

        self.assertIsNotNone(x_cluster.user)
        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_CHECK,
            name='cluster_check_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_check_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = orco.ClusterCheckRequest(identity='C1', params={'foo': 'bar'})

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_check,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'C1')

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_recover(self, notify, mock_find, mock_action):
        x_cluster = mock.Mock(id='CID')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterRecoverRequest(identity='C1',
                                         params={'operation': 'RECREATE'})

        result = self.eng.cluster_recover(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_RECOVER,
            name='cluster_recover_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'operation': 'RECREATE'},
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_recover_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = orco.ClusterRecoverRequest(identity='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_recover,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_recover_invalid(self, mock_find):
        x_cluster = mock.Mock(id='CID')
        mock_find.return_value = x_cluster

        req = orco.ClusterRecoverRequest(identity='Bogus',
                                         params={'bad': 'fake'})

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_recover,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Action parameter ['bad'] is not recognizable.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_recover_user_is_none(self, notify, mock_find,
                                          mock_action):
        x_cluster = mock.Mock(id='CID', project='PROJECT')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterRecoverRequest(identity='C1')

        result = self.eng.cluster_recover(self.ctx, req.obj_to_primitive())

        self.assertIsNotNone(x_cluster.user)
        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_RECOVER,
            name='cluster_recover_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes(self, mock_profile, mock_node):
        cluster = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.side_effect = [
            mock.Mock(id='OLD_ID', cluster_id='CID'),
            mock.Mock(id='NEW_ID', cluster_id='', status=consts.NS_ACTIVE,
                      profile_id='FAKE_ID_1')
        ]

        # do it
        res = self.eng._validate_replace_nodes(self.ctx, cluster,
                                               {'OLD_NODE': 'NEW_NODE'})

        self.assertEqual({'OLD_ID': 'NEW_ID'}, res)
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD_NODE'),
            mock.call(self.ctx, 'NEW_NODE'),
        ])
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'FAKE_ID', project_safe=True),
            mock.call(self.ctx, 'FAKE_ID_1', project_safe=True)
        ])

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_old_missing(self, mock_profile,
                                                 mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_node.side_effect = exc.ResourceNotFound(type='node', id='OLD')

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD': 'NEW'})

        self.assertIn("Original nodes not found: ['OLD']", six.text_type(ex))
        mock_node.assert_called_once_with(self.ctx, 'OLD')

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_new_missing(self, mock_profile,
                                                 mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_node.side_effect = [
            mock.Mock(),
            exc.ResourceNotFound(type='node', id='NEW')
        ]

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD': 'NEW'})

        self.assertIn("Replacement nodes not found: ['NEW']",
                      six.text_type(ex))
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD'),
            mock.call(self.ctx, 'NEW')
        ])

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_old_not_member(self, mock_profile,
                                                    mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_node.side_effect = [
            mock.Mock(cluster_id='OTHER'),
            mock.Mock(cluster_id=''),
        ]

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD': 'NEW'})

        self.assertIn("The specified nodes ['OLD'] to be replaced are not "
                      "members of the cluster CID.", six.text_type(ex))
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD'),
            mock.call(self.ctx, 'NEW')
        ])

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_new_not_orphan(self, mock_profile,
                                                    mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_node.side_effect = [
            mock.Mock(cluster_id='CID'),
            mock.Mock(cluster_id='OTHER'),
        ]

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD': 'NEW'})

        self.assertIn("Nodes ['NEW'] already member of a cluster.",
                      six.text_type(ex))
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD'),
            mock.call(self.ctx, 'NEW')
        ])

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_new_bad_status(self, mock_profile,
                                                    mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_node.side_effect = [
            mock.Mock(cluster_id='CID'),
            mock.Mock(cluster_id='', status=consts.NS_ERROR),
        ]

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD': 'NEW'})

        self.assertIn("Nodes are not ACTIVE: ['NEW'].", six.text_type(ex))
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD'),
            mock.call(self.ctx, 'NEW')
        ])

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_mult_err(self, mock_profile,
                                              mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_node.side_effect = [
            mock.Mock(id='OLD1', cluster_id='CID'),
            mock.Mock(id='NEW1', cluster_id='OTHER', status=consts.NS_ERROR),
        ]

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD1': 'NEW1'})

        msg1 = _("Nodes ['NEW1'] already member of a cluster.")
        msg2 = _("Nodes are not ACTIVE: ['NEW1'].")
        self.assertIn(msg1, six.text_type(ex))
        self.assertIn(msg2, six.text_type(ex))
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD1'),
            mock.call(self.ctx, 'NEW1'),
        ])

    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'get')
    def test__validate_replace_nodes_new_profile_type_mismatch(
            self, mock_profile, mock_node):
        c = mock.Mock(id='CID', profile_id='FAKE_CLUSTER_PROFILE')
        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE'),  # for cluster
            mock.Mock(type='FAKE_TYPE_1'),  # for node
        ]
        mock_node.side_effect = [
            mock.Mock(cluster_id='CID'),
            mock.Mock(cluster_id='', status=consts.NS_ACTIVE,
                      profile_id='FAKE_NODE_PROFILE'),
        ]

        # do it
        ex = self.assertRaises(exc.BadRequest,
                               self.eng._validate_replace_nodes,
                               self.ctx, c, {'OLD': 'NEW'})

        self.assertIn("Profile type of nodes ['NEW'] do not match that of "
                      "the cluster.", six.text_type(ex))
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'OLD'),
            mock.call(self.ctx, 'NEW')
        ])
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'FAKE_CLUSTER_PROFILE', project_safe=True),
            mock.call(self.ctx, 'FAKE_NODE_PROFILE', project_safe=True)
        ])

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, '_validate_replace_nodes')
    @mock.patch.object(no.Node, 'find')
    @mock.patch.object(po.Profile, 'find')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_replace_nodes(self, notify, mock_find,
                                   mock_profile, mock_node,
                                   mock_validate, mock_action):
        cluster = mock.Mock(id='CID', profile_id='FAKE_ID')
        mock_find.return_value = cluster
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        old_node = mock.Mock(id='ORIGIN', cluster_id='CID', status='ACTIVE')
        new_node = mock.Mock(id='REPLACE', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_ID_1')
        mock_node.side_effect = [old_node, new_node]
        mock_action.return_value = 'ACTION_ID'
        param = {'ORIGINAL': 'REPLACE'}
        mock_validate.return_value = param
        req = orco.ClusterReplaceNodesRequest(identity='CLUSTER', nodes=param)

        # do it
        res = self.eng.cluster_replace_nodes(self.ctx, req.obj_to_primitive())

        # verify
        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_validate.assert_called_once_with(self.ctx, cluster, param)
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_REPLACE_NODES,
            name='cluster_replace_nodes_CID',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'ORIGINAL': 'REPLACE'})
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, '_validate_replace_nodes')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_replace_nodes_failed_validate(self, mock_find, mock_chk):
        nodes = {'OLD': 'NEW'}
        cluster = mock.Mock()
        mock_find.return_value = cluster
        mock_chk.side_effect = exc.BadRequest(msg='failed')
        req = orco.ClusterReplaceNodesRequest(identity='CLUSTER', nodes=nodes)

        # do it
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_replace_nodes,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("failed.", six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_chk.assert_called_once_with(self.ctx, cluster, nodes)

    @mock.patch.object(nm.Node, 'load')
    @mock.patch.object(no.Node, 'get_all_by_cluster')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_collect(self, mock_find, mock_get, mock_load):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        x_obj_1 = mock.Mock(id='NODE1', physical_id='PHYID1')
        x_obj_1.to_dict.return_value = {'name': 'node1'}
        x_obj_2 = mock.Mock(id='NODE2', physical_id='PHYID2')
        x_obj_2.to_dict.return_value = {'name': 'node2'}
        x_node_1 = mock.Mock()
        x_node_2 = mock.Mock()
        x_node_1.get_details.return_value = {'ip': '1.2.3.4'}
        x_node_2.get_details.return_value = {'ip': '5.6.7.8'}
        mock_get.return_value = [x_obj_1, x_obj_2]
        mock_load.side_effect = [x_node_1, x_node_2]
        req = orco.ClusterCollectRequest(identity='CLUSTER_ID',
                                         path='details.ip')

        res = self.eng.cluster_collect(self.ctx, req.obj_to_primitive())

        self.assertIn('cluster_attributes', res)
        self.assertIn({'id': 'NODE1', 'value': '1.2.3.4'},
                      res['cluster_attributes'])
        self.assertIn({'id': 'NODE2', 'value': '5.6.7.8'},
                      res['cluster_attributes'])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_get.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_load.assert_has_calls([
            mock.call(self.ctx, db_node=x_obj_1),
            mock.call(self.ctx, db_node=x_obj_2)
        ])
        x_obj_1.to_dict.assert_called_once_with()
        x_node_1.get_details.assert_called_once_with(self.ctx)
        x_obj_2.to_dict.assert_called_once_with()
        x_node_2.get_details.assert_called_once_with(self.ctx)

    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(common_utils, 'get_path_parser')
    def test_cluster_collect_bad_path(self, mock_parser, mock_find):
        mock_parser.side_effect = exc.BadRequest(msg='Boom')
        req = orco.ClusterCollectRequest(identity='CLUSTER_ID', path='foo.bar')

        err = self.assertRaises(rpc.ExpectedException,
                                self.eng.cluster_collect,
                                self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, err.exc_info[0])
        mock_parser.assert_called_once_with('foo.bar')
        self.assertEqual(0, mock_find.call_count)

    @mock.patch.object(no.Node, 'get_all_by_cluster')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_collect_cluster_not_found(self, mock_find, mock_get):
        cid = 'FAKE_CLUSTER'
        mock_find.side_effect = exc.ResourceNotFound(type='cluster', id=cid)
        req = orco.ClusterCollectRequest(identity=cid, path='foo.bar')

        err = self.assertRaises(rpc.ExpectedException,
                                self.eng.cluster_collect,
                                self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, err.exc_info[0])
        mock_find.assert_called_once_with(self.ctx, cid)
        self.assertEqual(0, mock_get.call_count)

    @mock.patch.object(no.Node, 'get_all_by_cluster')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_collect_no_nodes(self, mock_find, mock_get):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        mock_get.return_value = []
        req = orco.ClusterCollectRequest(identity='CLUSTER_ID', path='barr')

        res = self.eng.cluster_collect(self.ctx, req.obj_to_primitive())

        self.assertEqual({'cluster_attributes': []}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_get.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')

    @mock.patch.object(no.Node, 'get_all_by_cluster')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_collect_no_details(self, mock_find, mock_get):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        x_node_1 = mock.Mock(id='NODE1', physical_id=None)
        x_node_1.to_dict.return_value = {'name': 'node1'}
        x_node_2 = mock.Mock(id='NODE2', physical_id=None)
        x_node_2.to_dict.return_value = {'name': 'node2'}
        mock_get.return_value = [x_node_1, x_node_2]
        req = orco.ClusterCollectRequest(identity='CLUSTER_ID', path='name')

        res = self.eng.cluster_collect(self.ctx, req.obj_to_primitive())

        self.assertIn('cluster_attributes', res)
        self.assertIn({'id': 'NODE1', 'value': 'node1'},
                      res['cluster_attributes'])
        self.assertIn({'id': 'NODE2', 'value': 'node2'},
                      res['cluster_attributes'])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_get.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        x_node_1.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_1.get_details.call_count)
        x_node_2.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_2.get_details.call_count)

    @mock.patch.object(no.Node, 'get_all_by_cluster')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_collect_no_match(self, mock_find, mock_get):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        x_node_1 = mock.Mock(physical_id=None)
        x_node_1.to_dict.return_value = {'name': 'node1'}
        x_node_2 = mock.Mock(physical_id=None)
        x_node_2.to_dict.return_value = {'name': 'node2'}
        mock_get.return_value = [x_node_1, x_node_2]
        req = orco.ClusterCollectRequest(identity='CLUSTER_ID', path='bogus')

        res = self.eng.cluster_collect(self.ctx, req.obj_to_primitive())

        self.assertEqual({'cluster_attributes': []}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_get.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        x_node_1.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_1.get_details.call_count)
        x_node_2.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_2.get_details.call_count)

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(ro.Receiver, 'get_all')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_delete(self, notify, mock_find, mock_policies,
                            mock_receivers, mock_action):
        x_obj = mock.Mock(id='12345678AB', status='ACTIVE', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = []
        mock_receivers.return_value = []
        mock_action.return_value = 'ACTION_ID'
        req = orco.ClusterDeleteRequest(identity='IDENTITY', force=False)

        result = self.eng.cluster_delete(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')
        mock_receivers.assert_called_once_with(
            self.ctx, filters={'cluster_id': '12345678AB'})
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_DELETE',
            name='cluster_delete_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY)

        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_delete_with_containers(self, mock_find):
        dependents = {'profiles': ['profile1']}
        cluster = mock.Mock(id='cluster1', status='ACTIVE',
                            dependents=dependents)
        mock_find.return_value = cluster
        req = orco.ClusterDeleteRequest(identity='FAKE_CLUSTER',
                                        force=False)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, req.obj_to_primitive())

        msg = _("The cluster 'FAKE_CLUSTER' cannot be deleted: still "
                "referenced by profile(s): ['profile1'].")
        self.assertEqual(exc.ResourceInUse, ex.exc_info[0])
        self.assertEqual(msg, six.text_type(ex.exc_info[1]))

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_delete_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')
        req = orco.ClusterDeleteRequest(identity='Bogus',
                                        force=False)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_delete_improper_status(self, mock_find):
        for bad_status in [consts.CS_CREATING, consts.CS_UPDATING,
                           consts.CS_DELETING, consts.CS_RECOVERING]:
            fake_cluster = mock.Mock(id='12345678AB', status=bad_status)
            mock_find.return_value = fake_cluster
            req = orco.ClusterDeleteRequest(identity='BUSY',
                                            force=False)

            ex = self.assertRaises(rpc.ExpectedException,
                                   self.eng.cluster_delete,
                                   self.ctx, req.obj_to_primitive())

            self.assertEqual(exc.ActionInProgress, ex.exc_info[0])
            self.assertEqual(
                "The cluster 'BUSY' is in status %s." % bad_status,
                six.text_type(ex.exc_info[1]))

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_delete_policy_attached(self, mock_find, mock_policies):
        x_obj = mock.Mock(id='12345678AB', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = [mock.Mock()]
        req = orco.ClusterDeleteRequest(identity='IDENTITY',
                                        force=False)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceInUse, ex.exc_info[0])
        expected_msg = _("The cluster 'IDENTITY' cannot be deleted: "
                         "there is still policy(s) attached to it.")
        self.assertEqual(expected_msg, six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')

    @mock.patch.object(ro.Receiver, 'get_all')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_delete_with_receiver(self, mock_find, mock_policies,
                                          mock_receivers):
        x_obj = mock.Mock(id='12345678AB', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = []
        mock_receivers.return_value = [mock.Mock()]
        req = orco.ClusterDeleteRequest(identity='IDENTITY',
                                        force=False)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceInUse, ex.exc_info[0])
        expected_msg = _("The cluster 'IDENTITY' cannot be deleted: "
                         "there is still receiver(s) associated with it.")
        self.assertEqual(expected_msg, six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')
        mock_receivers.assert_called_once_with(
            self.ctx, filters={'cluster_id': '12345678AB'})

    @mock.patch.object(ro.Receiver, 'get_all')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_delete_mult_err(self, mock_find, mock_policies,
                                     mock_receivers):
        x_obj = mock.Mock(id='12345678AB', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = [mock.Mock()]
        mock_receivers.return_value = [mock.Mock()]
        req = orco.ClusterDeleteRequest(identity='IDENTITY',
                                        force=False)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceInUse, ex.exc_info[0])
        self.assertIn('there is still policy(s) attached to it.',
                      six.text_type(ex.exc_info[1]))
        self.assertIn('there is still receiver(s) associated with it.',
                      six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')
        mock_receivers.assert_called_once_with(
            self.ctx, filters={'cluster_id': '12345678AB'})

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(ro.Receiver, 'get_all')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(co.Cluster, 'find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_delete_force(self, notify, mock_find, mock_policies,
                                  mock_receivers, mock_action):
        for bad_status in [consts.CS_CREATING, consts.CS_UPDATING,
                           consts.CS_DELETING, consts.CS_RECOVERING]:
            x_obj = mock.Mock(id='12345678AB', status=bad_status,
                              dependents={})
            mock_find.return_value = x_obj
            mock_policies.return_value = []
            mock_receivers.return_value = []
            mock_action.return_value = 'ACTION_ID'
            req = orco.ClusterDeleteRequest(identity='IDENTITY', force=True)

            result = self.eng.cluster_delete(self.ctx, req.obj_to_primitive())

            self.assertEqual({'action': 'ACTION_ID'}, result)
            mock_find.assert_called_with(self.ctx, 'IDENTITY')
            mock_policies.assert_called_with(self.ctx, '12345678AB')
            mock_receivers.assert_called_with(
                self.ctx, filters={'cluster_id': '12345678AB'})
            mock_action.assert_called_with(
                self.ctx, '12345678AB', 'CLUSTER_DELETE',
                name='cluster_delete_12345678',
                cause=consts.CAUSE_RPC,
                status=am.Action.READY)

            notify.assert_called_with()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.drivers.container import docker_v1
from senlin.tests.unit.common import base


class TestDocker(base.SenlinTestCase):

    @mock.patch("docker.APIClient")
    def setUp(self, mock_docker):
        super(TestDocker, self).setUp()
        self.x_docker = mock.Mock()
        mock_docker.return_value = self.x_docker
        self.sot = docker_v1.DockerClient("abc")

    @mock.patch("docker.APIClient")
    def test_init(self, mock_docker):
        x_docker = mock_docker.return_value
        url = mock.Mock()

        sot = docker_v1.DockerClient(url)

        self.assertEqual(x_docker, sot._dockerclient)
        mock_docker.assert_called_once_with(base_url=url, version='auto')

    def test_container_create(self):
        image = mock.Mock()

        self.sot.container_create(image)

        self.x_docker.create_container.assert_called_once_with(
            name=None, image=image, command=None)

    def test_container_delete(self):
        container = mock.Mock()

        res = self.sot.container_delete(container)

        self.assertTrue(res)
        self.x_docker.remove_container.assert_called_once_with(container)

    def test_restart(self):
        container = mock.Mock()

        res = self.sot.restart(container)

        self.assertIsNone(res)
        self.x_docker.restart.assert_called_once_with(container)

    def test_restart_with_wait(self):
        container = mock.Mock()

        res = self.sot.restart(container, timeout=20)

        self.assertIsNone(res)
        self.x_docker.restart.assert_called_once_with(container, timeout=20)

    def test_pause(self):
        container = mock.Mock()

        res = self.sot.pause(container)

        self.assertIsNone(res)
        self.x_docker.pause.assert_called_once_with(container)

    def test_unpause(self):
        container = mock.Mock()

        res = self.sot.unpause(container)

        self.assertIsNone(res)
        self.x_docker.unpause.assert_called_once_with(container)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
Client side of the senlin engine RPC API.
'''

from oslo_config import cfg

from senlin.common import consts
from senlin.common import messaging
from senlin.objects import base as object_base


class EngineClient(object):
    """Client side of the senlin engine rpc API.

    Version History:

      1.0 - Initial version (Mitaka 1.0 release)
      1.1 - Add cluster-collect call.
    """

    def __init__(self):
        serializer = object_base.VersionedObjectSerializer()
        self._client = messaging.get_rpc_client(consts.ENGINE_TOPIC,
                                                cfg.CONF.host,
                                                serializer=serializer)

    @staticmethod
    def make_msg(method, **kwargs):
        return method, kwargs

    def call(self, ctxt, method, req, version=None):
        """The main entry for invoking engine service.

        :param ctxt: The request context object.
        :param method: The name of the method to be invoked.
        :param req: A dict containing a request object.
        :param version: The engine RPC API version requested.
        """
        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client

        return client.call(ctxt, method, req=req)

    def cast(self, ctxt, msg, version=None):
        method, kwargs = msg
        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client
        return client.cast(ctxt, method, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.drivers import base
from senlin.drivers.openstack import sdk


class CeilometerClient(base.DriverBase):
    '''Ceilometer V2 driver.'''

    def __init__(self, params):
        super(CeilometerClient, self).__init__(params)
        self.conn = sdk.create_connection(params)

    @sdk.translate_exception
    def alarm_create(self, **attrs):
        return self.conn.telemetry.create_alarm(**attrs)

    @sdk.translate_exception
    def alarm_delete(self, value, ignore_missing=True):
        return self.conn.telemetry.delete_alarm(value, ignore_missing)

    @sdk.translate_exception
    def alarm_find(self, name_or_id, ignore_missing=True):
        return self.conn.telemetry.find_alarm(name_or_id, ignore_missing)

    @sdk.translate_exception
    def alarm_get(self, value):
        return self.conn.telemetry.get_alarm(value)

    @sdk.translate_exception
    def alarm_list(self, **query):
        return self.conn.telemetry.alarms(**query)

    @sdk.translate_exception
    def alarm_update(self, value, **attrs):
        return self.conn.telemetry.update_alarm(value, **attrs)

    @sdk.translate_exception
    def sample_create(self, **attrs):
        return self.conn.telemetry.create_sample(**attrs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg
from oslo_utils import timeutils
from oslo_utils import uuidutils

from senlin.common import exception as exc
from senlin.db import api as db_api
from senlin.objects import cluster as co
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestCluster(base.SenlinTestCase):

    def setUp(self):
        super(TestCluster, self).setUp()
        self.ctx = utils.dummy_context()

    @mock.patch.object(co.Cluster, 'get')
    def test_find_by_uuid(self, mock_get):
        x_cluster = mock.Mock()
        mock_get.return_value = x_cluster
        aid = uuidutils.generate_uuid()

        result = co.Cluster.find(self.ctx, aid)

        self.assertEqual(x_cluster, result)
        mock_get.assert_called_once_with(self.ctx, aid, project_safe=True)

    @mock.patch.object(co.Cluster, 'get_by_name')
    @mock.patch.object(co.Cluster, 'get')
    def test_find_by_uuid_as_name(self, mock_get, mock_get_name):
        x_cluster = mock.Mock()
        mock_get_name.return_value = x_cluster
        mock_get.return_value = None

        aid = uuidutils.generate_uuid()
        result = co.Cluster.find(self.ctx, aid, False)

        self.assertEqual(x_cluster, result)
        mock_get.assert_called_once_with(self.ctx, aid, project_safe=False)
        mock_get_name.assert_called_once_with(self.ctx, aid,
                                              project_safe=False)

    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_find_by_name(self, mock_get_name):
        x_cluster = mock.Mock()
        mock_get_name.return_value = x_cluster
        aid = 'this-is-not-uuid'

        result = co.Cluster.find(self.ctx, aid)

        self.assertEqual(x_cluster, result)
        mock_get_name.assert_called_once_with(self.ctx, aid, project_safe=True)

    @mock.patch.object(co.Cluster, 'get_by_short_id')
    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_find_by_shortid(self, mock_get_name, mock_get_shortid):
        x_cluster = mock.Mock()
        mock_get_shortid.return_value = x_cluster
        mock_get_name.return_value = None
        aid = 'abcd-1234-abcd'

        result = co.Cluster.find(self.ctx, aid, False)

        self.assertEqual(x_cluster, result)
        mock_get_name.assert_called_once_with(self.ctx, aid,
                                              project_safe=False)
        mock_get_shortid.assert_called_once_with(self.ctx, aid,
                                                 project_safe=False)

    @mock.patch.object(co.Cluster, 'get_by_short_id')
    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_find_not_found(self, mock_get_name, mock_get_short_id):
        mock_get_name.return_value = None
        mock_get_short_id.return_value = None

        self.assertRaises(exc.ResourceNotFound,
                          co.Cluster.find,
                          self.ctx, 'bogus')

        mock_get_name.assert_called_once_with(self.ctx, 'bogus',
                                              project_safe=True)
        mock_get_short_id.assert_called_once_with(self.ctx, 'bogus',
                                                  project_safe=True)

    @mock.patch.object(db_api, 'cluster_policy_ids_by_cluster')
    @mock.patch.object(db_api, 'node_ids_by_cluster')
    @mock.patch.object(db_api, 'profile_get')
    def test_to_dict(self, mock_profile, mock_nodes, mock_bindings):
        PROFILE_ID = '96f4df4b-889e-4184-ba8d-b5ca122f95bb'
        values = {
            'profile_id': PROFILE_ID,
            'name': 'test-cluster',
            'desired_capacity': 1,
            'status': 'INIT',
            'init_at': timeutils.utcnow(True),
            'max_size': -1,
            'min_size': 0,
            'timeout': cfg.CONF.default_action_timeout,
            'user': self.ctx.user,
            'project': self.ctx.project,
        }
        cluster = co.Cluster.create(self.ctx, values)
        fake_profile = mock.Mock()
        fake_profile.name = 'PROFILEABC'
        mock_profile.return_value = fake_profile
        mock_nodes.return_value = ['N1', 'N2']
        mock_bindings.return_value = ['P1', 'P2']
        expected = {
            'id': cluster.id,
            'name': cluster.name,
            'profile_id': PROFILE_ID,
            'user': cluster.user,
            'project': cluster.project,
            'domain': cluster.domain,
            'init_at': mock.ANY,
            'created_at': None,
            'updated_at': None,
            'min_size': 0,
            'max_size': -1,
            'desired_capacity': 1,
            'timeout': cfg.CONF.default_action_timeout,
            'status': 'INIT',
            'status_reason': None,
            'metadata': {},
            'data': {},
            'dependents': {},
            'config': {},
            'nodes': ['N1', 'N2'],
            'policies': ['P1', 'P2'],
            'profile_name': 'PROFILEABC',
        }

        self.assertEqual(expected, cluster.to_dict())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from tempest.lib import decorators
from tempest.lib import exceptions

from senlin.tests.tempest.api import base
from senlin.tests.tempest.common import constants


class TestProfileCreateNegativeBadRequest(base.BaseSenlinAPITest):

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('0f0183b8-0f5e-4367-993d-863ff1f30d49')
    def test_profile_create_profile_data_not_specified(self):
        params = {
            'profile': {
                'name': 'test-profile'
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'profiles', params)

        message = ex.resp_body['error']['message']
        self.assertEqual("Request body missing 'profile' key.",
                         str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('c341f22d-833b-4676-8b66-e6cdb0b77abd')
    def test_profile_create_name_not_specified(self):
        params = {
            'profile': {
                'spec': constants.spec_nova_server
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'profiles', params)

        message = ex.resp_body['error']['message']
        self.assertEqual("'name' is a required property", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('5e644149-a7e6-4e93-8220-4a32f98d6e25')
    def test_profile_create_spec_not_specified(self):
        params = {
            'profile': {
                'name': 'test-profile'
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'profiles', params)

        message = ex.resp_body['error']['message']
        self.assertEqual("'spec' is a required property", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('e2da6964-2cd2-402e-9004-ca6b7e3e63f1')
    def test_profile_create_invalid_param(self):
        params = {
            'profile': {
                'boo': 'foo'
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'profiles', params)

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "Additional properties are not allowed (u'boo' was "
            "unexpected)", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('591f3670-3fec-4645-bae2-4f6dec28d70c')
    def test_profile_create_profile_type_incorrect(self):
        spec = copy.deepcopy(constants.spec_nova_server)
        spec['type'] = 'senlin.profile.bogus'
        params = {
            'profile': {
                'name': 'test-profile',
                'spec': spec
            }
        }
        # Verify badrequest exception(404) is raised.
        ex = self.assertRaises(exceptions.NotFound,
                               self.client.create_obj,
                               'profiles', params)

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "The profile_type 'senlin.profile.bogus-1.0' could "
            "not be found.", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('66977f7a-5d30-481c-a5ec-a445e80a7c0f')
    def test_profile_create_spec_validation_failed(self):
        spec = copy.deepcopy(constants.spec_nova_server)
        spec['properties']['bogus'] = 'foo'
        params = {
            'profile': {
                'name': 'test-profile',
                'spec': spec
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'profiles', params)

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "Failed in creating profile test-profile: "
            "Unrecognizable spec item 'bogus'", str(message))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import decorators
from tempest.lib import exceptions

from senlin.tests.tempest.api import base
from senlin.tests.tempest.common import utils


class TestClusterCreateNegativeBadRequest(base.BaseSenlinAPITest):

    def setUp(self):
        super(TestClusterCreateNegativeBadRequest, self).setUp()
        profile_id = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, profile_id)

        self.profile_id = profile_id

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('498a06eb-8c5f-4d9e-852f-87ac295f1a96')
    def test_cluster_create_cluster_data_not_specified(self):
        # cluster key is missing in request data
        params = {
            'bad': {
                'profile_id': self.profile_id,
                'name': 'test-cluster'
            }
        }

        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'clusters', params)

        message = ex.resp_body['error']['message']
        self.assertEqual("Request body missing 'cluster' key.",
                         str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('3aced30b-ccb2-4e40-90c2-7b6aa205e3d6')
    def test_cluster_create_profile_invalid(self):
        # Invalid profile_id is provided
        params = {
            'cluster': {
                'profile_id': '3aced30b-ccb2-4e40-90c2-7b6aa205e3d6',
                'name': 'test-cluster'
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'clusters', params)

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "The specified profile '3aced30b-ccb2-4e40-90c2-7b6aa205e3d6' "
            "could not be found.", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('61b190bb-ef5a-47b3-acbe-6467fbb44439')
    def test_cluster_create_miss_profile(self):
        # Invalid profile_id is provided
        params = {
            'cluster': {
                'name': 'test-cluster'
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'clusters', params)

        message = ex.resp_body['error']['message']
        self.assertEqual("'profile_id' is a required property",
                         str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('7eaf60c3-f33d-403b-a4ee-0276ae90928c')
    def test_cluster_create_size_constraint_illegal(self):
        # Invalid size limitation is defined: min_size > max_size
        desired_capacity = 2
        min_size = 5
        max_size = 3
        params = {
            'cluster': {
                'profile_id': self.profile_id,
                'desired_capacity': desired_capacity,
                'min_size': min_size,
                'max_size': max_size,
                'name': 'test-cluster'
            }
        }

        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.create_obj,
                               'clusters', params)

        message = ex.resp_body['error']['message']
        self.assertEqual("The target capacity (2) is less than the "
                         "specified min_size (5).", str(message))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from oslo_serialization import jsonutils
import six
from six.moves import urllib
import yaml

from senlin.common.i18n import _


# Try LibYAML if available
if hasattr(yaml, 'CSafeLoader'):
    Loader = yaml.CSafeLoader
else:
    Loader = yaml.SafeLoader

if hasattr(yaml, 'CSafeDumper'):
    Dumper = yaml.CSafeDumper
else:
    Dumper = yaml.SafeDumper


class YamlLoader(Loader):
    def normalise_file_path_to_url(self, path):
        if urllib.parse.urlparse(path).scheme:
            return path
        path = os.path.abspath(path)
        return urllib.parse.urljoin('file:',
                                    urllib.request.pathname2url(path))

    def include(self, node):
        try:
            url = self.normalise_file_path_to_url(self.construct_scalar(node))
            tmpl = urllib.request.urlopen(url).read()
            return yaml.safe_load(tmpl)
        except urllib.error.URLError as ex:
            raise IOError('Failed retrieving file %s: %s' %
                          (url, six.text_type(ex)))

    def process_unicode(self, node):
        # Override the default string handling function to always return
        # unicode objects
        return self.construct_scalar(node)


YamlLoader.add_constructor('!include', YamlLoader.include)
YamlLoader.add_constructor(u'tag:yaml.org,2002:str',
                           YamlLoader.process_unicode)
YamlLoader.add_constructor(u'tag:yaml.org,2002:timestamp',
                           YamlLoader.process_unicode)


def simple_parse(in_str):
    try:
        out_dict = jsonutils.loads(in_str)
    except ValueError:
        try:
            out_dict = yaml.load(in_str, Loader=YamlLoader)
        except yaml.YAMLError as yea:
            yea = six.text_type(yea)
            msg = _('Error parsing input: %s') % yea
            raise ValueError(msg)
        else:
            if out_dict is None:
                out_dict = {}

    if not isinstance(out_dict, dict):
        msg = _('The input is not a JSON object or YAML mapping.')
        raise ValueError(msg)

    return out_dict



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import inspect
import re
import six
import webob

from oslo_config import cfg
from oslo_log import log
from oslo_messaging._drivers import common as rpc_common

import senlin.api.middleware.fault as fault
from senlin.common import exception as senlin_exc
from senlin.tests.unit.common import base


class ClusterNotFoundChild(senlin_exc.ResourceNotFound):
    pass


class ErrorWithNewline(webob.exc.HTTPBadRequest):
    pass


class FaultMiddlewareTest(base.SenlinTestCase):
    def setUp(self):
        super(FaultMiddlewareTest, self).setUp()
        log.register_options(cfg.CONF)

    def test_disguised_http_exception_with_newline(self):
        wrapper = fault.FaultWrapper(None)
        newline_error = ErrorWithNewline('Error with \n newline')
        msg = wrapper._error(senlin_exc.HTTPExceptionDisguise(newline_error))
        expected = {
            'code': 400,
            'error': {
                'code': 400,
                'message': 'Error with \n newline',
                'type': 'ErrorWithNewline'
            },
            'explanation': 'The server could not comply with the request '
                           'since it is either malformed or otherwise '
                           'incorrect.',
            'title': 'Bad Request'
        }
        self.assertEqual(expected, msg)

    def test_openstack_exception_with_kwargs(self):
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(senlin_exc.ResourceNotFound(type='cluster',
                                                         id='a'))

        expected = {
            "code": 404,
            "error": {
                "code": 404,
                "message": "The cluster 'a' could not be found.",
                "type": "ResourceNotFound"
            },
            "explanation": "The resource could not be found.",
            "title": "Not Found"
        }
        self.assertEqual(expected, msg)

    def test_openstack_exception_without_kwargs(self):
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(senlin_exc.PolicyNotSpecified())
        expected = {
            'code': 500,
            'error': {
                'code': 500,
                'message': 'Policy not specified.',
                'type': 'PolicyNotSpecified'
            },
            'explanation': 'The server has either erred or is incapable of '
                           'performing the requested operation.',
            'title': 'Internal Server Error'
        }
        self.assertEqual(expected, msg)

    def test_exception_with_non_ascii_chars(self):
        # We set debug to true to test the code path for serializing traces too
        cfg.CONF.set_override('debug', True)
        msg = u'Error with non-ascii chars \x80'

        class TestException(senlin_exc.SenlinException):
            msg_fmt = msg

        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(TestException())

        self.assertEqual(500, msg['code'])
        self.assertEqual(500, msg['error']['code'])
        self.assertEqual(u'Error with non-ascii chars \x80',
                         msg['error']['message'])
        self.assertEqual('TestException', msg['error']['type'])
        self.assertEqual('The server has either erred or is incapable of '
                         'performing the requested operation.',
                         msg['explanation'])
        self.assertEqual('Internal Server Error', msg['title'])

    def test_remote_exception(self):
        cfg.CONF.set_override('debug', True)
        error = senlin_exc.ResourceNotFound(type='cluster', id='a')
        exc_info = (type(error), error, None)
        serialized = rpc_common.serialize_remote_exception(exc_info)
        remote_error = rpc_common.deserialize_remote_exception(
            serialized, ["senlin.common.exception"])
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(remote_error)
        expected_message = six.text_type(remote_error).split('\n', 1)[0]
        expected = {
            'code': 404,
            'error': {
                'code': 404,
                'message': expected_message,
                'type': 'ResourceNotFound'
            },
            'explanation': 'The resource could not be found.',
            'title': 'Not Found'
        }
        self.assertEqual(expected, msg)

    def remote_exception_helper(self, name, error):
        exc_info = (type(error), error, None)

        serialized = rpc_common.serialize_remote_exception(exc_info)
        remote_error = rpc_common.deserialize_remote_exception(
            serialized, name)
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(remote_error)
        expected = {
            'code': 500,
            'error': {
                'code': 500,
                'message': msg['error']['message'],
                'type': 'RemoteError'
            },
            'explanation': msg['explanation'],
            'title': 'Internal Server Error'
        }
        self.assertEqual(expected, msg)

    def test_all_remote_exceptions(self):
        for name, obj in inspect.getmembers(
                senlin_exc, lambda x: inspect.isclass(x) and issubclass(
                    x, senlin_exc.SenlinException)):

            if '__init__' in obj.__dict__:
                if obj == senlin_exc.SenlinException:
                    continue
                elif obj == senlin_exc.Error:
                    error = obj('Error')
                elif obj == senlin_exc.ResourceNotFound:
                    error = obj()
                else:
                    continue
                self.remote_exception_helper(name, error)
                continue

            if hasattr(obj, 'msg_fmt'):
                kwargs = {}
                spec_names = re.findall('%\((\w+)\)([cdeEfFgGinorsxX])',
                                        obj.msg_fmt)

                for key, convtype in spec_names:
                    if convtype == 'r' or convtype == 's':
                        kwargs[key] = '"' + key + '"'
                    else:
                        # this is highly unlikely
                        raise Exception("test needs additional conversion"
                                        " type added due to %s exception"
                                        " using '%c' specifier" % (
                                            obj, convtype))

                error = obj(**kwargs)
                self.remote_exception_helper(name, error)

    def test_should_not_ignore_parent_classes(self):
        wrapper = fault.FaultWrapper(None)

        msg = wrapper._error(ClusterNotFoundChild(type='cluster', id='a'))
        expected = {
            "code": 404,
            "error": {
                "code": 404,
                "message": "The cluster 'a' could not be found.",
                "type": "ClusterNotFoundChild"
            },
            "explanation": "The resource could not be found.",
            "title": "Not Found"
        }
        self.assertEqual(expected, msg)

    def test_internal_server_error_when_exception_and_parents_not_mapped(self):
        wrapper = fault.FaultWrapper(None)

        class NotMappedException(Exception):
            pass

        msg = wrapper._error(NotMappedException('A message'))
        expected = {
            "code": 500,
            "error": {
                "code": 500,
                "message": "A message",
                "type": "NotMappedException"
            },
            "explanation": ("The server has either erred or is incapable "
                            "of performing the requested operation."),
            "title": "Internal Server Error"
        }
        self.assertEqual(expected, msg)

    def test_should_not_ignore_parent_classes_even_for_remote_ones(self):
        cfg.CONF.set_override('debug', True)

        error = ClusterNotFoundChild(type='cluster', id='a')
        exc_info = (type(error), error, None)
        serialized = rpc_common.serialize_remote_exception(exc_info)
        remote_error = rpc_common.deserialize_remote_exception(
            serialized, ["senlin.tests.unit.api.middleware.test_fault"])

        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(remote_error)
        expected_message = six.text_type(remote_error).split('\n', 1)[0]
        expected = {
            'code': 404,
            'error': {
                'code': 404,
                'message': expected_message,
                'type': 'ClusterNotFoundChild'
            },
            'explanation': 'The resource could not be found.',
            'title': 'Not Found'
        }
        self.assertEqual(expected, msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import decorators
from tempest.lib import exceptions

from senlin.tests.tempest.api import base
from senlin.tests.tempest.common import utils


class TestPolicyUpdateNegativeNotFound(base.BaseSenlinAPITest):

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('5df90d82-9889-4c6f-824c-30272bcfa767')
    def test_policy_update_policy_not_found(self):
        ex = self.assertRaises(exceptions.NotFound,
                               self.client.update_obj, 'policies',
                               '5df90d82-9889-4c6f-824c-30272bcfa767',
                               {'policy': {'name': 'new-name'}})

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "The policy '5df90d82-9889-4c6f-824c-30272bcfa767' "
            "could not be found.", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('29414add-9cba-4b72-a7bb-36718671dcab')
    def test_policy_update_policy_invalid_param(self):
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.update_obj, 'policies',
                               '5df90d82-9889-4c6f-824c-30272bcfa767',
                               {'policy': {'boo': 'foo'}})

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "Additional properties are not allowed (u'boo' was "
            "unexpected)", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('bf26ed1e-1d26-4472-b4c8-0bcca1c0a838')
    def test_policy_update_policy_empty_param(self):
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.update_obj, 'policies',
                               '5df90d82-9889-4c6f-824c-30272bcfa767',
                               {})

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "Malformed request data, missing 'policy' key in "
            "request body.", str(message))


class TestPolicyUpdateNegativeBadRequest(base.BaseSenlinAPITest):

    def setUp(self):
        super(TestPolicyUpdateNegativeBadRequest, self).setUp()
        # Create a policy
        policy_id = utils.create_a_policy(self)
        self.addCleanup(utils.delete_a_policy, self, policy_id)
        self.policy_id = policy_id

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('31242de5-55ac-4589-87a1-a9940e4beca2')
    def test_policy_update_no_property_updated(self):
        # No property is updated.
        params = {
            'policy': {}
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.update_obj,
                               'policies', self.policy_id, params)

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "'name' is a required property", str(message))

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('d2ca7de6-0069-48c9-b3de-ee975a2428dc')
    def test_policy_update_spec_not_updatable(self):
        # Try to update spec of policy.
        # Note: name is the only property that can be updated
        # after policy is created.
        params = {
            'policy': {
                'name': 'new-name',
                'spec': {'k1': 'v1'}
            }
        }
        # Verify badrequest exception(400) is raised.
        ex = self.assertRaises(exceptions.BadRequest,
                               self.client.update_obj,
                               'policies', self.policy_id, params)

        message = ex.resp_body['error']['message']
        self.assertEqual(
            "Additional properties are not allowed (u'spec' was "
            "unexpected)", str(message))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import decorators

from senlin.tests.tempest.common import constants
from senlin.tests.tempest.common import utils
from senlin.tests.tempest.functional import base


class TestBatchPolicy(base.BaseSenlinFunctionalTest):

    def setUp(self):
        super(TestBatchPolicy, self).setUp()
        self.new_profile2 = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, self.new_profile2)
        self.new_profile = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, self.new_profile)
        self.old_profile = utils.create_a_profile(self)
        self.addCleanup(utils.delete_a_profile, self, self.old_profile)
        self.cluster_id = utils.create_a_cluster(self, self.old_profile,
                                                 min_size=0, max_size=2,
                                                 desired_capacity=2)
        self.addCleanup(utils.delete_a_cluster, self, self.cluster_id)

    @decorators.attr(type=['functional'])
    @decorators.idempotent_id('b35e63ed-f1d5-4ec6-afb0-2378d260afcb')
    def test_batch_policy(self):
        # Create a batch policy targets on CLUSTER_UPDATE action
        spec = constants.spec_batch_policy
        policy_id = utils.create_a_policy(self, spec)
        self.addCleanup(utils.delete_a_policy, self, policy_id)

        # Attach batch policy to cluster
        utils.cluster_attach_policy(self, self.cluster_id, policy_id)
        self.addCleanup(utils.cluster_detach_policy, self,
                        self.cluster_id, policy_id)

        utils.update_a_cluster(self, self.cluster_id,
                               profile_id=self.new_profile)

        # Verify cluster update result
        cluster = utils.get_a_cluster(self, self.cluster_id)
        self.assertEqual('ACTIVE', cluster['status'])
        self.assertEqual(2, len(cluster['nodes']))
        self.assertEqual(self.new_profile, cluster['profile_id'])
